@@ -12,6 +12,8 @@ import logging
 import sqlalchemy
 import io
 from datetime import datetime
+
+from commcare_export.checkpoint import CheckpointManager
 from six.moves import input
 
 import dateutil.parser
@@ -24,6 +26,7 @@ from commcare_export.commcare_minilinq import CommCareHqEnv
 from commcare_export import writers
 from commcare_export import excel_query
 from commcare_export import misc
+from commcare_export.version import __version__
 
 logger = logging.getLogger(__name__)
 
@@ -33,23 +36,27 @@ commcare_hq_aliases = {
 }
 
 def main(argv):
-    parser = argparse.ArgumentParser('commcare-hq-export', 'Output a customized export of CommCareHQ data.')
+    parser = argparse.ArgumentParser('commcare-export', 'Output a customized export of CommCareHQ data.')
 
+    parser.add_argument('--version', default=False, action='store_true', help='Print the current version of the commcare-export tool.')
     parser.add_argument('--query', help='JSON or Excel query file. If omitted, JSON string is read from stdin.')
     parser.add_argument('--dump-query', default=False, action='store_true')
-    parser.add_argument('--commcare-hq', default='prod')
+    parser.add_argument('--commcare-hq', default='prod', help='Base url for the CommCare HQ instance e.g. https://www.commcarehq.org')
     parser.add_argument('--api-version', default=LATEST_KNOWN_VERSION)
-    parser.add_argument('--project', required=True)
+    parser.add_argument('--project')
     parser.add_argument('--username')
     parser.add_argument('--password')
-    parser.add_argument('--auth-mode', default='session', help='Use "session" based auth or "digest" auth.')
-    parser.add_argument('--since')
-    parser.add_argument('--start-over', default=False, action='store_true', 
+    parser.add_argument('--auth-mode', default='digest', help='Use "session" based auth or "digest" auth.')
+    parser.add_argument('--since', help='Export all data after this date. Format YYYY-MM-DD or YYYY-MM-DDTHH:mm:SS')
+    parser.add_argument('--until', help='Export all data up until this date. Format YYYY-MM-DD or YYYY-MM-DDTHH:mm:SS')
+    parser.add_argument('--start-over', default=False, action='store_true',
                         help='When saving to a SQL database; the default is to pick up since the last success. This disables that.')
     parser.add_argument('--profile')
     parser.add_argument('--verbose', default=False, action='store_true')
     parser.add_argument('--output-format', default='json', choices=['json', 'csv', 'xls', 'xlsx', 'sql', 'markdown'], help='Output format')
     parser.add_argument('--output', metavar='PATH', default='reports.zip', help='Path to output; defaults to `reports.zip`.')
+    parser.add_argument('--strict-types', default=False, action='store_true', help="When saving to a SQL database don't allow changing column types once they are created.")
+    parser.add_argument('--missing-value', default=None, help="Value to use when a field is missing from the form / case.")
 
     args = parser.parse_args(argv)
 
@@ -60,6 +67,15 @@ def main(argv):
         logging.basicConfig(level=logging.WARN,
                             format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
 
+    if args.version:
+        print('commcare-export version {}'.format(__version__))
+        exit(0)
+
+    if not args.project:
+        print('commcare-export: error: argument --project is required')
+        exit(1)
+
+
     if args.profile:
         # hotshot is gone in Python 3
         import hotshot
@@ -84,13 +100,17 @@ def main_with_args(args):
     
     # Reads as excel if it is a file name that looks like excel, otherwise reads as JSON, 
     # falling back to parsing arg directly as JSON, and finally parsing stdin as JSON
+    missing_value = args.missing_value
+    if args.output_format != 'sql' and missing_value is None:
+        missing_value = ''
+
     if args.query:
         if os.path.exists(args.query):
             query_file_md5 = misc.digest_file(args.query)
             if os.path.splitext(args.query)[1] in ['.xls', '.xlsx']:
                 import openpyxl
                 workbook = openpyxl.load_workbook(args.query)
-                query = excel_query.compile_workbook(workbook)
+                query = excel_query.compile_workbook(workbook, missing_value)
             else:
                 with io.open(args.query, encoding='utf-8') as fh:
                     query = MiniLinq.from_jvalue(json.loads(fh.read()))
@@ -98,7 +118,13 @@ def main_with_args(args):
             print('Query file not found: %s' % args.query)
             exit(1)
     else:
-        query = MiniLinq.from_jvalue(json.loads(sys.stdin.read()))
+        try:
+            query = MiniLinq.from_jvalue(json.loads(sys.stdin.read()))
+        except Exception as e:
+            raise Exception(
+                "Failure reading query from console input. "
+                "Try using the '--query' parameter to pass your query as an Excel file", e
+            )
 
     if args.dump_query:
         print(json.dumps(query.to_jvalue(), indent=4))
@@ -112,17 +138,22 @@ def main_with_args(args):
         args.password = getpass.getpass()
 
     # Build an API client using either the URL provided, or the URL for a known alias
-    api_client = CommCareHqClient(url = commcare_hq_aliases.get(args.commcare_hq, args.commcare_hq), 
+    commcarehq_base_url = commcare_hq_aliases.get(args.commcare_hq, args.commcare_hq)
+    api_client = CommCareHqClient(url =commcarehq_base_url,
                                   project = args.project,
                                   version = args.api_version)
 
     api_client = api_client.authenticated(username=args.username, password=args.password, mode=args.auth_mode)
-
+    checkpoint_manager = None
     if args.output_format == 'xlsx':
         writer = writers.Excel2007TableWriter(args.output)
     elif args.output_format == 'xls':
         writer = writers.Excel2003TableWriter(args.output)
     elif args.output_format == 'csv':
+        if not args.output.endswith(".zip"):
+            print("WARNING: csv output is a zip file, but "
+                  "will be written to %s" % args.output)
+            print("Consider appending .zip to the file name to avoid confusion.")
         writer = writers.CsvTableWriter(args.output)
     elif args.output_format == 'json':
         writer = writers.JValueTableWriter()
@@ -133,29 +164,28 @@ def main_with_args(args):
         # Writer had bizarre issues so we use a full connection instead of passing in a URL or engine
         import sqlalchemy
         engine = sqlalchemy.create_engine(args.output)
-        writer = writers.SqlTableWriter(engine.connect())
+        is_mysql = 'mysql' in args.output
+        collation = 'utf8_bin' if is_mysql else None
+        writer = writers.SqlTableWriter(engine.connect(), args.strict_types, collation=collation)
+        checkpoint_manager = CheckpointManager(engine.connect(), collation=collation)
+        with checkpoint_manager:
+            checkpoint_manager.create_checkpoint_table()
+        api_client.set_checkpoint_manager(checkpoint_manager, query=args.query, query_md5=query_file_md5)
 
         if not args.since and not args.start_over and os.path.exists(args.query):
-            connection = sqlalchemy.create_engine(args.output)
+            with checkpoint_manager:
+                args.since = checkpoint_manager.get_time_of_last_run(query_file_md5)
 
-            # Grab the current list of tables to see if we have already run & written to it
-            metadata = sqlalchemy.MetaData()
-            metadata.bind = connection
-            metadata.reflect()
-
-            if 'commcare_export_runs' in metadata.tables:
-                cursor = connection.execute(sqlalchemy.sql.text('SELECT time_of_run FROM commcare_export_runs WHERE query_file_md5 = :query_file_md5 ORDER BY time_of_run DESC'), query_file_md5=query_file_md5)
-                for row in cursor:
-                    args.since = row[0]
-                    logger.debug('Last successful run was %s', args.since)
-                    break
-                cursor.close()
+            if args.since:
+                logger.debug('Last successful run was %s', args.since)
             else:
                 logger.warn('No successful runs found, and --since not specified: will import ALL data')
 
     if args.since:
         logger.debug('Starting from %s', args.since)
-    env = BuiltInEnv() | CommCareHqEnv(api_client, since=dateutil.parser.parse(args.since) if args.since else None) | JsonPathEnv({}) 
+    since = dateutil.parser.parse(args.since) if args.since else None
+    until = dateutil.parser.parse(args.until) if args.until else None
+    env = BuiltInEnv({'commcarehq_base_url': commcarehq_base_url}) | CommCareHqEnv(api_client, since=since, until=until) | JsonPathEnv({})
     results = query.eval(env)
 
     # Assume that if any tables were emitted, that is the idea, otherwise print the output
@@ -163,14 +193,16 @@ def main_with_args(args):
         with writer:
             for table in env.emitted_tables():
                 logger.debug('Writing %s', table['name'])
+                if table['name'] != table['name'].lower():
+                    logger.warning(
+                        "Caution: Using upper case letters in a "
+                        "table name is not advised: {}".format(table['name'])
+                    )
                 writer.write_table(table)
 
-            if args.output_format == 'sql' and os.path.exists(args.query):
-                writer.write_table({
-                    'name': 'commcare_export_runs',
-                    'headings': ['id', 'query_file_name', 'query_file_md5', 'time_of_run'],
-                    'rows': [ [uuid.uuid4().hex, args.query, query_file_md5, run_start.isoformat()] ]
-                })
+        if checkpoint_manager and os.path.exists(args.query):
+            with checkpoint_manager:
+                checkpoint_manager.set_checkpoint(args.query, query_file_md5, run_start, True)
 
         if args.output_format == 'json':
             print(json.dumps(writer.tables, indent=4, default=RepeatableIterator.to_jvalue))
