@@ -1,10 +1,13 @@
 from __future__ import unicode_literals, print_function, absolute_import, division, generators, nested_scopes
+from datetime import datetime
 import operator
+import pytz
 import six
 from itertools import chain
 
 from jsonpath_rw import jsonpath
 from jsonpath_rw.parser import parse as parse_jsonpath
+from commcare_export.misc import unwrap, unwrap_val
 
 from commcare_export.repeatable_iterator import RepeatableIterator
 
@@ -133,11 +136,11 @@ class DictEnv(Env):
         self.d = d or {}
 
     def bind(self, name, value):
-        return dict(self.d.items() + [(name, value)])
+        return DictEnv(dict(list(self.d.items()) + [(name, value)]))
         
     def lookup(self, name):
         try:             return self.d[name]
-        except KeyError: raise NotFound()
+        except KeyError: raise NotFound(unwrap_val(name))
 
     def replace(self, data):
         if isinstance(data, dict): return DictEnv(data)
@@ -178,7 +181,7 @@ class JsonPathEnv(Env):
         elif isinstance(name, jsonpath.JSONPath):
             jsonpath_expr = name
         else:
-            raise NotImplementedError() 
+            raise NotFound(unwrap_val(name))
 
         def iter(jsonpath_expr=jsonpath_expr): # Capture closure
             for datum in jsonpath_expr.find(self.__bindings):
@@ -217,6 +220,139 @@ class JsonPathEnv(Env):
 # Actual concrete environments, basically with built-in functions.
 #
 
+
+def _not_val(val):
+    return val is None or val == []
+
+
+def _to_unicode(val):
+    if isinstance(val, bytes):
+        return val.decode('utf8')
+    elif not isinstance(val, six.text_type):
+        return six.text_type(val)
+
+    return val
+
+
+@unwrap('val')
+def str2bool(val):
+    if _not_val(val):
+        return False
+
+    if isinstance(val, bool):
+        return val
+
+    val = _to_unicode(val)
+    return bool(val) and val.lower() in {'true', 't', '1'}
+
+
+@unwrap('val')
+def str2num(val):
+    if _not_val(val):
+        return None
+
+    try:
+        return int(val)
+    except ValueError:
+        return float(val)
+
+
+@unwrap('val')
+def str2date(val):
+    import dateutil.parser as parser
+    if not val:
+        return None
+
+    val = _to_unicode(val)
+
+    try:
+        date = parser.parse(val)
+    except ValueError:
+        return
+
+    try:
+        date = date.astimezone(pytz.utc)
+    except ValueError:
+        pass
+
+    return date.replace(microsecond=0, tzinfo=None)
+
+@unwrap('val')
+def bool2int(val):
+    return int(str2bool(val))
+
+
+@unwrap('val')
+def selected_at(val, index):
+    if not val:
+        return None
+
+    try:
+        index = int(index)
+    except ValueError:
+        return "Error: index must be an integer"
+
+    val = _to_unicode(val)
+
+    try:
+        return val.split()[index]
+    except (IndexError, ValueError):
+        return None
+
+
+@unwrap('val')
+def selected(val, reference):
+    if not val:
+        return None
+
+    val = _to_unicode(val)
+
+    parts = val.split()
+    return reference in parts
+
+
+@unwrap('val')
+def count_selected(val):
+    if not val:
+        return None
+
+    val = _to_unicode(val)
+
+    return len(val.split())
+
+
+def join(*args):
+    args = [unwrap_val(arg)for arg in args]
+    return args[0].join(args[1:])
+
+
+@unwrap('val')
+def default(val, default_val):
+    if not val:
+        return default_val
+    return val
+
+
+@unwrap('val')
+def attachment_url(val):
+    if not val:
+        return None
+    from commcare_export.minilinq import Apply, Reference, Literal
+    return Apply(
+        Reference('template'),
+        Literal('{}/a/{}/api/form/attachment/{}/{}'),
+        Reference('commcarehq_base_url'),
+        Reference('$.domain'),
+        Reference('$.id'),
+        Literal(val)
+    )
+
+
+def template(format_template, *args):
+    args = [unwrap_val(arg) for arg in args]
+    return format_template.format(*args)
+
+
 class BuiltInEnv(DictEnv):
     """
     A built-in environment of operators and functions
@@ -227,20 +363,34 @@ class BuiltInEnv(DictEnv):
     the first env involved in almost any situation.
     """
     
-    def __init__(self):
+    def __init__(self, d=None):
         self.__tables = []
-        return super(BuiltInEnv, self).__init__({
-            '+'   : operator.__add__,
-            '-'   : operator.__sub__,
-            '*'   : operator.__mul__,
-            '//'  : operator.__floordiv__,
-            '/'   : operator.__truediv__,
-            '>'   : operator.__gt__,
-            '>='  : operator.__ge__,
-            '<'   : operator.__lt__,
-            '<='  : operator.__le__,
-            'len' : len,
+        d = d or {}
+        d.update({
+            '+': operator.__add__,
+            '-': operator.__sub__,
+            '*': operator.__mul__,
+            '//': operator.__floordiv__,
+            '/': operator.__truediv__,
+            '>': operator.__gt__,
+            '>=': operator.__ge__,
+            '<': operator.__lt__,
+            '<=': operator.__le__,
+            'len': len,
+            'bool': bool,
+            'str2bool': str2bool,
+            'bool2int': bool2int,
+            'str2num': str2num,
+            'str2date': str2date,
+            'selected': selected,
+            'selected-at': selected_at,
+            'count-selected': count_selected,
+            'join': join,
+            'default': default,
+            'template': template,
+            'attachment_url': attachment_url,
         })
+        return super(BuiltInEnv, self).__init__(d)
 
     def bind(self, name, value): raise CannotBind()
     def replace(self, data): raise CannotReplace()
