@@ -1,15 +1,13 @@
-import re
+import csv
+import datetime
+import logging
 import sys
 import zipfile
-import csv
-import json
-import logging
+from itertools import chain
 
 import six
 from six import StringIO, u
 
-from itertools import chain
-
 logger = logging.getLogger(__name__)
 
 MAX_COLUMN_SIZE = 2000
@@ -19,6 +17,20 @@ def ensure_text(v):
         return v
     elif isinstance(v, six.binary_type):
         return u(v)
+    elif isinstance(v, datetime.datetime):
+        return v.strftime('%Y-%m-%d %H:%M:%S')
+    elif isinstance(v, datetime.date):
+        return v.isoformat()
+    elif v is None:
+        return ''
+    else:
+        return u(str(v))
+
+def to_jvalue(v):
+    if isinstance(v, (six.text_type,) + six.integer_types):
+        return v
+    elif isinstance(v, six.binary_type):
+        return u(v)
     else:
         return u(str(v))
 
@@ -36,7 +48,7 @@ class TableWriter(object):
     
     def write_table(self, table):
         "{'name': str, 'headings': [str], 'rows': [[str]]} -> ()"
-        raise NotImplementedError() 
+        raise NotImplementedError()
 
     def __exit__(self, exc_type, exc_val, exc_tb):
         pass
@@ -141,7 +153,7 @@ class JValueTableWriter(TableWriter):
         # Ensures the table is iterable; probably better to create a custom JSON handler that runs in constant space
         self.tables.append(dict(name=table['name'],
                                 headings=list(table['headings']),
-                                rows=[list(row) for row in table['rows']]))
+                                rows=[[to_jvalue(v) for v in row] for row in table['rows']]))
 
 class StreamingMarkdownTableWriter(TableWriter):
     """
@@ -156,18 +168,19 @@ class StreamingMarkdownTableWriter(TableWriter):
         self.output_stream.write('|%s|\n' % '|'.join(table['headings']))
 
         for row in table['rows']:
-            self.output_stream.write('|%s|\n' % '|'.join(row))
+            self.output_stream.write('|%s|\n' % '|'.join(ensure_text(val) for val in row))
 
-class SqlTableWriter(TableWriter):
+
+class SqlMixin(object):
     """
     Write tables to a database specified by URL
     (TODO) with "upsert" based on primary key.
     """
 
-    MIN_VARCHAR_LEN=32  # Since SQLite does not actually support ALTER COLUMN type, let's maximize the chance that we do not have to write workarounds by starting medium
-    MAX_VARCHAR_LEN=255 # Arbitrary point at which we switch to TEXT; for postgres VARCHAR == TEXT anyhow and for Sqlite it doesn't matter either
+    MIN_VARCHAR_LEN = 32  # Since SQLite does not actually support ALTER COLUMN type, let's maximize the chance that we do not have to write workarounds by starting medium
+    MAX_VARCHAR_LEN = 255  # Arbitrary point at which we switch to TEXT; for postgres VARCHAR == TEXT anyhow and for Sqlite it doesn't matter either
 
-    def __init__(self, connection):
+    def __init__(self, connection, collation=None):
         try:
             import sqlalchemy
             import alembic
@@ -179,10 +192,11 @@ class SqlTableWriter(TableWriter):
                             "command:  pip install sqlalchemy alembic")
 
         self.base_connection = connection
+        self.collation = collation
 
     def __enter__(self):
-        self.connection = self.base_connection.connect() # "forks" the SqlAlchemy connection
-        return self # TODO: fork the writer so this can be called many times
+        self.connection = self.base_connection.connect()  # "forks" the SqlAlchemy connection
+        return self  # TODO: fork the writer so this can be called many times
 
     def __exit__(self, exc_type, exc_val, exc_tb):
         self.connection.close()
@@ -202,7 +216,37 @@ class SqlTableWriter(TableWriter):
     def table(self, table_name):
         return self.sqlalchemy.Table(table_name, self.metadata, autoload=True, autoload_with=self.connection)
 
+    def get_id_column(self):
+        return self.sqlalchemy.Column(
+            'id',
+            self.sqlalchemy.Unicode(self.MAX_VARCHAR_LEN),
+            primary_key=True
+        )
+
+
+class SqlTableWriter(SqlMixin, TableWriter):
+    """
+    Write tables to a database specified by URL
+    (TODO) with "upsert" based on primary key.
+    """
+
+    def __init__(self, connection, strict_types=False, collation=None):
+        super(SqlTableWriter, self).__init__(connection, collation=collation)
+        self.strict_types = strict_types
+
+    @property
+    def is_sqllite(self):
+        return 'sqlite' in self.connection.engine.driver
+
     def best_type_for(self, val):
+        if not self.is_sqllite:
+            if isinstance(val, bool):
+                return self.sqlalchemy.Boolean()
+            elif isinstance(val, datetime.datetime):
+                return self.sqlalchemy.DateTime()
+            elif isinstance(val, datetime.date):
+                return self.sqlalchemy.Date()
+
         if isinstance(val, int):
             return self.sqlalchemy.Integer()
         elif isinstance(val, six.string_types):
@@ -213,27 +257,21 @@ class SqlTableWriter(TableWriter):
             # 3. But SQLite really barfs on altering columns. Luckily it does actually have real types,
             #    so we count on other parts of this code to not bother running column alterations
             if len(val) < self.MAX_VARCHAR_LEN: # FIXME: Is 255 an interesting cutoff?
-                return self.sqlalchemy.Unicode( max(len(val), self.MIN_VARCHAR_LEN) )
+                return self.sqlalchemy.Unicode( max(len(val), self.MIN_VARCHAR_LEN), collation=self.collation)
             else:
-                return self.sqlalchemy.UnicodeText()
+                return self.sqlalchemy.UnicodeText(collation=self.collation)
         else:
             # We do not have a name for "bottom" in SQL aka the type whose least upper bound
             # with any other type is the other type.
-            return None
+            return self.sqlalchemy.UnicodeText(collation=self.collation)
 
     def compatible(self, source_type, dest_type):
         """
         Checks _coercion_ compatibility.
         """
-
-        # FIXME: Add datetime and friends
-        if isinstance(source_type, self.sqlalchemy.Integer):
-            # Integers can be cast to varch
-            return True
-
         if isinstance(source_type, self.sqlalchemy.String):
             if not isinstance(dest_type, self.sqlalchemy.String):
-                False
+                return False
             elif source_type.length is None:
                 # The length being None means that we are looking at indefinite strings aka TEXT.
                 # This tool will never create strings with bounds, but if a target DB has one then
@@ -241,7 +279,19 @@ class SqlTableWriter(TableWriter):
                 # We will request that whomever uses this tool convert to TEXT type.
                 return dest_type.length is None
             else:
-                return (dest_type.length >= source_type.length)
+                return dest_type.length is None or (dest_type.length >= source_type.length)
+
+        compatibility = {
+            self.sqlalchemy.String: (self.sqlalchemy.Text,),
+            self.sqlalchemy.Integer: (self.sqlalchemy.String, self.sqlalchemy.Text),
+            self.sqlalchemy.Boolean: (self.sqlalchemy.String, self.sqlalchemy.Text, self.sqlalchemy.Integer),
+            self.sqlalchemy.DateTime: (self.sqlalchemy.String, self.sqlalchemy.Text, self.sqlalchemy.Date),
+            self.sqlalchemy.Date: (self.sqlalchemy.String, self.sqlalchemy.Text),
+        }
+        for _type, types in compatibility.items():
+            if isinstance(source_type, _type):
+                return isinstance(dest_type, (_type,) + types)
+
 
     def least_upper_bound(self, source_type, dest_type):
         """
@@ -251,48 +301,80 @@ class SqlTableWriter(TableWriter):
         """
 
         # FIXME: Don't be so silly
-        return self.sqlalchemy.UnicodeText()
+        return self.sqlalchemy.UnicodeText(collation=self.collation)
 
     def make_table_compatible(self, table_name, row_dict):
-        # FIXME: This does lots of redundant checks in a tight loop. Stop doing that.
-        
         ctx = self.alembic.migration.MigrationContext.configure(self.connection)
         op = self.alembic.operations.Operations(ctx)
 
         if not table_name in self.metadata.tables:
-            id_column = self.sqlalchemy.Column(
-                'id',
-                self.sqlalchemy.Unicode(self.MAX_VARCHAR_LEN),
-                primary_key=True
-            )
-            op.create_table(table_name, id_column)
+            def get_columns():
+                return [self.get_id_column()] + [
+                    self.sqlalchemy.Column(name, self.best_type_for(val), nullable=True)
+                    for name, val in row_dict.items() if val is not None and name != 'id'
+                ]
+
+            if self.strict_types:
+                create_sql = self.sqlalchemy.schema.CreateTable(self.sqlalchemy.Table(
+                    table_name,
+                    self.sqlalchemy.MetaData(),
+                    *get_columns()
+                )).compile(self.connection.engine)
+                logger.warn("Table '{table_name}' does not exist. Creating table with:\n{schema}".format(
+                    table_name=table_name,
+                    schema=create_sql
+                ))
+                empty_cols = [name for name, val in row_dict.items() if val is None]
+                if empty_cols:
+                    logger.warn("This schema does not include the following columns since we are unable "
+                                "to determine the column type at this stage: {}".format(empty_cols))
+            op.create_table(table_name, *get_columns())
+            self.metadata.clear()
             self.metadata.reflect()
+            return
+
+        def get_cols():
+            return {c.name: c for c in self.table(table_name).columns}
+
+        columns = get_cols()
 
         for column, val in row_dict.items():
-            ty = self.best_type_for(val)
+            if val is None:
+                continue
 
-            if not column in [c.name for c in self.table(table_name).columns]:
-                # If we are creating the column, a None crashes things even though it is the "empty" type
-                # but SQL does not have such a type. So we have to guess a liberal type for future use.
-                ty = ty or self.sqlalchemy.UnicodeText()
+            ty = self.best_type_for(val)
+            if not column in columns:
+                logger.warn("Adding column '{}.{} {}'".format(table_name, column, ty))
                 op.add_column(table_name, self.sqlalchemy.Column(column, ty, nullable=True))
                 self.metadata.clear()
                 self.metadata.reflect()
-
+                columns = get_cols()
             else:
-                columns = dict([(c.name, c) for c in self.table(table_name).columns])
+                if self.strict_types:
+                    # don't bother checking compatibility since we're not going to change anything
+                    continue
+
                 current_ty = columns[column].type
 
-                if not self.compatible(ty, current_ty) and not ('sqlite' in self.connection.engine.driver):
-                    op.alter_column(table_name, column, type_ = self.least_upper_bound(current_ty, ty))
+                if not self.compatible(ty, current_ty):
+                    new_type = self.least_upper_bound(ty, current_ty)
+                    if self.is_sqllite:
+                        logger.warn('Type mismatch detected for column %s (%s != %s) '
+                                    'but sqlite does not support changing column types', columns[column], current_ty, new_type)
+                        continue
+                    logger.warn('Altering column %s from %s to %s for value: "%s:%s"', columns[column], current_ty, new_type, type(val), val)
+                    op.alter_column(table_name, column, type_=new_type)
                     self.metadata.clear()
                     self.metadata.reflect()
+                    columns = get_cols()
 
     def upsert(self, table, row_dict):
-
         # For atomicity "insert, catch, update" is slightly better than "select, insert or update".
         # The latter may crash, while the former may overwrite data (which should be fine if whatever is
         # racing against this is importing from the same source... if not you are busted anyhow
+
+        # strip out values that are None since the column may not exist yet
+        row_dict = {col: val for col, val in row_dict.items() if val is not None}
         try:
             insert = table.insert().values(**row_dict)
             self.connection.execute(insert)
@@ -301,6 +383,9 @@ class SqlTableWriter(TableWriter):
             self.connection.execute(update)
 
     def write_table(self, table):
+        """
+        :param table: dict of {'name': 'name', 'headings', [...], 'rows': [[...], [...]]
+        """
         table_name = table['name']
         headings = table['headings']
 
@@ -313,4 +398,3 @@ class SqlTableWriter(TableWriter):
             self.upsert(self.table(table_name), row_dict)
 
         if logger.getEffectiveLevel() == 'DEBUG': sys.stderr.write('\n')
-        
